/*
   MyideDisk - base class for accessing a MyIDE drive

   Copyright (C) 2006 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "MyideDisk.h"
#include "MyidePartitionInfoCHS.h"
#include "MyidePartitionInfoLBA.h"

#include "Error.h"

MyideDisk::MyideDisk(RCPtr<RawDisk> rawdisk)
	: super(),
	  fRawDisk(rawdisk)
{
}

MyideDisk::~MyideDisk()
{
}

bool MyideDisk::ReadPartitionTable()
{
	unsigned char buf[256];

	if (!ReadBlock256(0, buf)) {
		return false;
	}
	unsigned char* id = buf + MyidePartitionInfo::eOffsetDriveNumbersPartitionVersion;

	// buf[4] is 0..15 in case of old CHS partition
	if (*id <= 15) {
		fPartitionInfo = new MyidePartitionInfoCHS();
		return fPartitionInfo->ParsePartitionInfo(buf);
	}
	if (*id == MyidePartitionInfo::eLBAPartitionTableMagic_old || 
		(id[0] == MyidePartitionInfo::eLBAPartitionTableMagic_0 &&
		 id[1] == MyidePartitionInfo::eLBAPartitionTableMagic_1 &&
		 id[2] == MyidePartitionInfo::eLBAPartitionTableMagic_2
		)
	) {
		RCPtr<MyidePartitionInfoLBA> info = new MyidePartitionInfoLBA();
		fPartitionInfo = info;
		if (!info->ParsePartitionInfo(buf)) {
			return false;
		}
		unsigned int max = info->GetPartitionTableSectors();
		for (unsigned int sec = 1; sec <= max; sec ++) {
			if (!ReadBlock256(sec, buf)) {
				return false;
			}
			if (!info->ParseAdditionalPartitonSector(sec, buf)) {
				return false;
			}
		}
		return true;
	}
	throw ErrorObject("Unknown partition table format");
	return false;

}

bool MyideDisk::WritePartitionTable()
{
	unsigned char buf[256];

	if (!fPartitionInfo->CreatePartitionInfo(buf)) {
		return false;
	}

	if (!WriteBlock256(0, buf)) {
		return false;
	}
	if (fPartitionInfo->IsLbaPartitionTable()) {
		RCPtr<MyidePartitionInfoLBA> info = RCPtrStaticCast<MyidePartitionInfoLBA>(fPartitionInfo);
		unsigned int max = info->GetPartitionTableSectors();
		for (unsigned int sec = 1; sec <= max; sec ++) {
			if (!info->CreateAdditionalPartitonSector(sec, buf)) {
				return false;
			}
			if (!WriteBlock256(sec, buf)) {
				return false;
			}
		}
	}
	return true;
}


bool MyideDisk::CreateDefaultLbaPartitionTable()
{
	RCPtr<MyidePartitionInfoLBA> part = new MyidePartitionInfoLBA();
	fPartitionInfo = part;

	unsigned int i;

	unsigned int max_sectors = fRawDisk->GetDiskSizeInSectors();

	if (max_sectors == 0) {
		return false;
	}

	printf("detected disk size: %d sectors (%d MB)\n", max_sectors, max_sectors / 2048);

	//if (max_sectors < 2 + 721 + 65536*2 + 721*2) {
	if (max_sectors < 2 + 721*2) {
		printf("disk too small - cannot create default partition table\n");
		return false;
	}

	part->SetDiskSize(max_sectors);

	part->SetPartitionTableSectors(1);
	part->SetPartitionTableVersion(0x10);

	unsigned int sector = 2;
	unsigned int partno = 0;

	for (i=1; i<= fPartitionInfo->GetNumberOfPartitions(); i++) {
		switch(i) {
		case 2:
			//part->SetPartition(i, true, true, i+2, sector, 65536, MyidePartitionInfoLBA::ePartTypeQD);
			part->SetPartition(i, true, true, i+2, sector, 721, MyidePartitionInfoLBA::ePartTypeQD);
			sector += 721;
			partno++;
			break;
		case 1:
			//part->SetPartition(i, true, true, i+2, sector, 65536*2, MyidePartitionInfoLBA::ePartTypeQD_emulated);
			part->SetPartition(i, true, true, i+2, sector, 721*2, MyidePartitionInfoLBA::ePartTypeQD_emulated);
			sector += 721*2;
			partno++;
			break;
/*
		case 2:
			part->SetPartition(i, true, true, i+2, sector, 65536, eDD);
			sector += 65536;
			partno++;
			break;
		case 3:
			part->SetPartition(i, true, true, i+2, sector, 721, eSD);
			sector += 721;
			partno++;
			break;
		case 4:
			part->SetPartition(i, true, true, i+2, sector, 721, eDD);
			sector += 721;
			partno++;
			break;
*/
/*
		case 16:
			part->SetPartition(i, true, true, 0, sector, 721, eDD);
			sector += 721;
			break;
*/
		default:
/*
			if (i <= 31) {
				part->SetPartition(i, true, true, 0, sector, 721, MyidePartitionInfoLBA::ePartTypeDD);
				partno++;
				sector += 721;
			} else {
*/
				part->SetPartition(i, false, false, 0, 0, 0, MyidePartitionInfoLBA::ePartTypeNone);
//			}
			break;
		}
	}
/*
	unsigned int image_slots = (max_sectors - sector) / 1041;
	if (image_slots > 0) {
		part->SetImageSpace(true, true, sector, image_slots * 1041);
	} else {
		part->SetImageSpace(false, false, 0, 0);
	}
*/

	part->SetOptionByte(0);
	part->SetPartitionsNo(partno);
	part->SetBootNo(3);

	return true;
}

bool MyideDisk::ReadBlock512(unsigned int sector, unsigned char* buf)
{
	Assert(buf);
	if (!fRawDisk->ReadBlock(sector, buf)) {
		return false;
	}
	return true;
}

bool MyideDisk::ReadBlock256(unsigned int sector, unsigned char* buf)
{
	Assert(buf);
	if (!fRawDisk->ReadBlock(sector, fSectorBuffer)) {
		return false;
	}
	unsigned int i;
	for (i=0;i<256;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::ReadBlock128(unsigned int sector, unsigned char* buf)
{
	Assert(buf);
	if (!fRawDisk->ReadBlock(sector, fSectorBuffer)) {
		return false;
	}
	unsigned int i;
	for (i=0;i<128;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::WriteBlock512(unsigned int sector, const unsigned char* buf)
{
	Assert(buf);
	if (!fRawDisk->WriteBlock(sector, buf)) {
		return false;
	}
	return true;
}

bool MyideDisk::WriteBlock256(unsigned int sector, const unsigned char* buf)
{
	Assert(buf);
	unsigned int i;
	memset(fSectorBuffer, 0, 512);
	for (i=0;i<256;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}
	if (!fRawDisk->WriteBlock(sector, fSectorBuffer)) {
		return false;
	}
	return true;
}

bool MyideDisk::WriteBlock128(unsigned int sector, const unsigned char* buf)
{
	Assert(buf);
	unsigned int i;
	memset(fSectorBuffer, 0, 512);
	for (i=0;i<128;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}
	if (!fRawDisk->WriteBlock(sector, fSectorBuffer)) {
		return false;
	}
	return true;
}


bool MyideDisk::CalculateImageSector(unsigned int imageno, unsigned int sector, unsigned int& diskSector) const
{
	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	if (!IsValidImageSector(sector)) {
		return false;
	}

	diskSector = fPartitionInfo->GetImageSectorNumber(imageno, sector);
	return true;
}

bool MyideDisk::ReadImageSector(unsigned int imageno, unsigned int sector, EDensity density, unsigned char* buf)
{
	unsigned int diskSector;

	if (!CalculateImageSector(imageno, sector, diskSector)) {
		return false;
	}
	switch (density) {
	case eSD:
	case eED:
		return ReadBlock128(diskSector, buf);
	case eDD:
		if (sector <= 3) {
			return ReadBlock128(diskSector, buf);
		} else {
			return ReadBlock256(diskSector, buf);
		}
	case eQD:
		// all sectors are 512 bytes
		return ReadBlock512(diskSector, buf);
	default:
		break;
	}
	return false;
}

bool MyideDisk::WriteImageSector(unsigned int imageno, unsigned int sector, EDensity density, const unsigned char* buf)
{
	unsigned int diskSector;

	if (!CalculateImageSector(imageno, sector, diskSector)) {
		return false;
	}
	switch (density) {
	case eSD:
	case eED:
		return WriteBlock128(diskSector, buf);
	case eDD:
		if (sector <= 3) {
			return WriteBlock128(diskSector, buf);
		} else {
			return WriteBlock256(diskSector, buf);
		}
	case eQD:
		// all sectors are 512 bytes
		return WriteBlock512(diskSector, buf);
	default:
		break;
	}
	return false;
}

bool MyideDisk::CalculatePartitionSector(unsigned int partno, unsigned int sector, unsigned int& diskSector) const
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	if (!fPartitionInfo->IsActivePartition(partno)) {
		return false;
	}
	if (!IsValidPartitionSector(partno, sector)) {
		return false;
	}

	diskSector = fPartitionInfo->GetPartitionSectorNumber(partno, sector);
	return true;
}

bool MyideDisk::ReadPartitionSector(unsigned int partno, unsigned int sector, EDensity density, unsigned char* buf)
{
	unsigned int diskSector;

	if (!CalculatePartitionSector(partno, sector, diskSector)) {
		return false;
	}
	switch (density) {
	case eSD:
	case eED:
		return ReadBlock128(diskSector, buf);
	case eDD:
		if (sector <= 3) {
			return ReadBlock128(diskSector, buf);
		} else {
			return ReadBlock256(diskSector, buf);
		}
	case eQD:
		if (fPartitionInfo->IsEmulatedQDPartition(partno)) {
			if (!ReadBlock256(diskSector, buf)) {
				return false;
			}
			return ReadBlock256(diskSector+1, buf+256);
		} else {
			return ReadBlock512(diskSector, buf);
		}
	default:
		break;
	}
	return false;
}

bool MyideDisk::WritePartitionSector(unsigned int partno, unsigned int sector, EDensity density, const unsigned char* buf)
{
	unsigned int diskSector;

	if (!CalculatePartitionSector(partno, sector, diskSector)) {
		return false;
	}
	switch (density) {
	case eSD:
	case eED:
		return WriteBlock128(diskSector, buf);
	case eDD:
		if (sector <= 3) {
			return WriteBlock128(diskSector, buf);
		} else {
			return WriteBlock256(diskSector, buf);
		}
	case eQD:
		if (fPartitionInfo->IsEmulatedQDPartition(partno)) {
			if (!WriteBlock256(diskSector, buf)) {
				return false;
			}
			return WriteBlock256(diskSector+1, buf+256);
		} else {
			return WriteBlock512(diskSector, buf);
		}
	default:
		break;
	}
	return false;
}

bool MyideDisk::CalculateImageInfoSector(unsigned int imageno, unsigned int& diskSector) const
{
	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	diskSector = fPartitionInfo->GetImageInfoSectorNumber(imageno);
	return true;
}

bool MyideDisk::ClearImageInfo(unsigned int imageno)
{
	unsigned char buf[512];
	unsigned int sector;

	if (!CalculateImageInfoSector(imageno, sector)) {
		return false;
	}

	memset(buf, 0, 512);

	if (!WriteBlock512(sector, buf)) {
		return false;
	}
	return true;
}


bool MyideDisk::GetImageInfo(unsigned int imageno, MyideImage& img)
{
	unsigned char buf[256];
	unsigned int sector;

	if (!CalculateImageInfoSector(imageno, sector)) {
		return false;
	}

	if (!ReadBlock256(sector, buf)) {
		return false;
	}

	int i=0,j=0;
	char c;
	while (i<eMyideImageNameLen) {
		c = buf[i] & 0x7f;
		i++;
		if (c < 64) {
			c += 32;
		} else if (c < 96) {
			c -= 64;
		}
		if (c < 32 || c == 96 || c == 123 || c >= 125) {
			continue;
		}
		img.Name[j++] = c;
	}
	img.Name[j] = 0;
	j--;
	while (j >= 0 && img.Name[j] == 32) {
		img.Name[j--] = 0;
	}
	switch (buf[eMyideDensityOffset]) {
	case 0xa4:
		img.Density = eDD; break;
	case 0xad:
		img.Density = eED; break;
	case 0xb3:
		img.Density = eSD; break;
	default:
		img.Density = eUnknownDensity; break;
	}
	return true;
}

bool MyideDisk::SetImageInfo(unsigned int imageno, const MyideImage& img)
{
	unsigned char buf[256];
	unsigned int sector;

	if (!CalculateImageInfoSector(imageno, sector)) {
		return false;
	}

	memset(buf, 0, 256);

	switch (img.Density) {
	case eSD: buf[eMyideDensityOffset] = 0xb3; break;
	case eED: buf[eMyideDensityOffset] = 0xad; break;
	case eDD: buf[eMyideDensityOffset] = 0xa4; break;
	default: return false;
	}

	int i=0,j=0;
	char c;
	while (i < eMyideImageNameLen && j < eMyideImageNameLen && (c=img.Name[i])) {
		i++;
		if (c < 32 || c == 96 || c == 123 || c >= 125) {
			continue;
		}
		if (c < 32) {
			c += 64;
		} else if (c < 96) {
			c -= 32;
		}
		buf[j++] = c;
	}

	if (!WriteBlock256(sector, buf)) {
		return false;
	}
	return true;
}

bool MyideDisk::ConvertCylinderToSectorNumber(unsigned int cylinder, unsigned int& sector) const
{
	if (!IsChsPartitionTable()) {
		return false;
	}
	RCPtr<MyidePartitionInfoCHS> pi = RCPtrStaticCast<MyidePartitionInfoCHS>(fPartitionInfo);
	sector = pi->GetNumberOfSectorsPerCylinder() * cylinder;
	return true;
}

bool MyideDisk::ConvertSectorNumberToCylinder(unsigned int sector, unsigned int& cylinder) const
{
	if (!IsChsPartitionTable()) {
		return false;
	}
	RCPtr<MyidePartitionInfoCHS> pi = RCPtrStaticCast<MyidePartitionInfoCHS>(fPartitionInfo);
	cylinder = sector / pi->GetNumberOfSectorsPerCylinder(); 
	return true;
}
