#ifndef MYIDEDISK_H
#define MYIDEDISK_H

/*
   MyideDisk - base class for accessing a MyIDE drive

   Copyright (C) 2006 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "RefCounted.h"
#include "RCPtr.h"
#include "AtariDebug.h"
#include "MyidePartitionInfo.h"
#include "RawDisk.h"

enum { eMyideImageNameLen = 29,
       eMyideDensityOffset = 30 };

typedef struct {
	EDensity Density;
	char Name[eMyideImageNameLen+1];
} MyideImage;

class MyideDisk : public RefCounted {
public:
	MyideDisk(RCPtr<RawDisk> rawdisk);

	virtual ~MyideDisk();

	bool ReadPartitionTable();
	bool WritePartitionTable();

	bool CreateDefaultLbaPartitionTable();

	inline bool IsChsPartitionTable() const;
	inline bool IsLbaPartitionTable() const;

	// partition functions

	inline void PrintPartitionTable() const;

	inline unsigned int GetNumberOfPartitions() const;

	inline bool IsActivePartition(unsigned int partno) const;
	inline unsigned int GetPartitionSize(unsigned int num) const;
	inline EDensity GetPartitionDensity(unsigned int num) const;

	inline bool IsValidPartitionNumber(unsigned int partno) const;

	bool ReadPartitionSector(unsigned int partno, unsigned int sector, EDensity density, unsigned char* buf);
	bool WritePartitionSector(unsigned int partno, unsigned int sector, EDensity density, const unsigned char* buf);


	// image functions

	inline bool ImageSpaceEnabled() const;
	inline unsigned int GetNumberOfImageSlots() const;
	inline unsigned int GetSectorsPerImageSlot() const;

	// convenience function: check if 1 <= imgno <= NumberOfImageSlots
	inline bool IsValidImageNumber(unsigned int imgno) const;

	bool ClearImageInfo(unsigned int imageno);
	bool GetImageInfo(unsigned int imageno, MyideImage& img);
	bool SetImageInfo(unsigned int imageno, const MyideImage& img);

	bool ReadImageSector(unsigned int imageno, unsigned int sector, EDensity density, unsigned char* buf);
	bool WriteImageSector(unsigned int imageno, unsigned int sector, EDensity density, const unsigned char* buf);

	// direct disk access - be careful!
	bool ReadBlock128(unsigned int sector, unsigned char* buf);
	bool ReadBlock256(unsigned int sector, unsigned char* buf);
	bool ReadBlock512(unsigned int sector, unsigned char* buf);
	bool WriteBlock128(unsigned int sector, const unsigned char* buf);
	bool WriteBlock256(unsigned int sector, const unsigned char* buf);
	bool WriteBlock512(unsigned int sector, const unsigned char* buf);

	// only valid for old partition format!
	bool ConvertCylinderToSectorNumber(unsigned int cylinder, unsigned int& sector) const;
	bool ConvertSectorNumberToCylinder(unsigned int sector, unsigned int& cylinder) const;

private:
	typedef RefCounted super;


	inline bool IsValidImageSector(unsigned int secno) const;

	inline bool IsValidPartitionSector(unsigned int partno, unsigned int secno) const;

	inline bool IsValidSectorLength(unsigned int sector, EDensity density, unsigned int len);

	// check if parameters are valid and calculate absolute disk sector
	bool CalculatePartitionSector(unsigned int imageno, unsigned int sector, unsigned int& diskSector) const;
	bool CalculateImageSector(unsigned int imageno, unsigned int sector, unsigned int& diskSector) const;
	bool CalculateImageInfoSector(unsigned int imageno, unsigned int& diskSector) const;

	RCPtr<RawDisk> fRawDisk;
	RCPtr<MyidePartitionInfo> fPartitionInfo;

	mutable unsigned char fSectorBuffer[512];
};

inline bool MyideDisk::IsChsPartitionTable() const
{
	return fPartitionInfo->IsChsPartitionTable();
}

inline bool MyideDisk::IsLbaPartitionTable() const
{
	return fPartitionInfo->IsLbaPartitionTable();
}

inline void MyideDisk::PrintPartitionTable() const
{
	fPartitionInfo->PrintPartitionTable();
}

inline unsigned int MyideDisk::GetNumberOfPartitions() const
{
	return fPartitionInfo->GetNumberOfPartitions();
}

inline bool MyideDisk::IsActivePartition(unsigned int partno) const
{
	return fPartitionInfo->IsActivePartition(partno);
}

inline unsigned int MyideDisk::GetPartitionSize(unsigned int num) const
{
	return fPartitionInfo->GetPartitionSize(num);
}

inline EDensity MyideDisk::GetPartitionDensity(unsigned int num) const
{
	return fPartitionInfo->GetPartitionDensity(num);
}

inline bool MyideDisk::IsValidPartitionNumber(unsigned int partno) const
{
	return fPartitionInfo->IsValidPartitionNumber(partno);
}

inline bool MyideDisk::IsValidPartitionSector(unsigned int partno, unsigned int secno) const
{
	if (!IsValidPartitionNumber(partno)) {
		Assert(false);
		return false;
	}
	return fPartitionInfo->IsValidPartitionSector(partno, secno);
}

inline bool MyideDisk::ImageSpaceEnabled() const
{
	return fPartitionInfo->ImageSpaceEnabled();
}

inline unsigned int MyideDisk::GetNumberOfImageSlots() const
{
	return fPartitionInfo->GetNumberOfImageSlots();
}

inline unsigned int MyideDisk::GetSectorsPerImageSlot() const
{
	return fPartitionInfo->GetSectorsPerImageSlot();
}

inline bool MyideDisk::IsValidImageNumber(unsigned int imgno) const
{
	return (imgno >= 1) && (imgno <= GetNumberOfImageSlots());
}

inline bool MyideDisk::IsValidImageSector(unsigned int secno) const
{
	return fPartitionInfo->IsValidImageSector(secno);
}

inline bool MyideDisk::IsValidSectorLength(unsigned int sector, EDensity density, unsigned int seclen)
{
	return fPartitionInfo->IsValidSectorLength(sector, density, seclen);
}

#endif
