/*
   MyidePartitionInfoLBA - infos for (new) LBA partition table

   Copyright (C) 2008-2011 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "MyidePartitionInfoLBA.h"
#include "MiscUtils.h"

MyidePartitionInfoLBA::MyidePartitionInfoLBA()
	: fPartitionInfo(0),
	  fTotalLBASectors(0),
	  fImageSpaceAvailable(false),
	  fImageSpaceSlots(0),
	  fMaxPartitions(0),
          fPartitions(0),
          fBootNo(0),
	  fPtableSectors(0),
	  fPtableVersion(0)
{ }

MyidePartitionInfoLBA::~MyidePartitionInfoLBA()
{
	if (fPartitionInfo) {
		delete[] fPartitionInfo;
		fPartitionInfo = 0;
	}
}

bool MyidePartitionInfoLBA::SetPartitionTableSectors(unsigned int sectors)
{
	if (fPartitionInfo) {
		delete[] fPartitionInfo;
		fPartitionInfo = 0;
	}
	if (sectors > eMaxPtableSectors) {
		fMaxPartitions = 0;
		return false;
	}
	fPtableSectors = sectors;
	fMaxPartitions = 15 + (fPtableSectors * 16);
	fPartitionInfo = new LBAPartition[fMaxPartitions];
	return true;
}

bool MyidePartitionInfoLBA::ParsePartitionInfo(const unsigned char* buf)
{
	const unsigned char* id = buf + MyidePartitionInfo::eOffsetDriveNumbersPartitionVersion;

	if (*id == MyidePartitionInfo::eLBAPartitionTableMagic_old) {
		if (!SetPartitionTableSectors(buf[eOfsPtableSectors_old])) {
			return false;
		}

		fTotalLBASectors = buf[eOfsTotalLBASize_old]
			+ (buf[eOfsTotalLBASize_old + 1] << 8)
			+ (buf[eOfsTotalLBASize_old + 2] << 16)
			+ (buf[eOfsTotalLBASize_old + 3] << 24);

		fOptionByte = buf[eOfsOptionByte_old];
		fPartitions = buf[eOfsPartitions_old];
		fBootNo = buf[eOfsBootNo_old];
		fPtableVersion = 0x09;

		fImageSpaceAvailable = false;
		fImageSpaceSlots = 0;

		return ParseAdditionalPartitonSector(0, buf);
	} else if (id[0] == MyidePartitionInfo::eLBAPartitionTableMagic_0 &&
		id[1] == MyidePartitionInfo::eLBAPartitionTableMagic_1 &&
		id[2] == MyidePartitionInfo::eLBAPartitionTableMagic_2) {

		if (!SetPartitionTableSectors(buf[eOfsPtableSectors])) {
			return false;
		}

		fTotalLBASectors = buf[eOfsTotalLBASize]
			+ (buf[eOfsTotalLBASize + 1] << 8)
			+ (buf[eOfsTotalLBASize + 2] << 16)
			+ (buf[eOfsTotalLBASize + 3] << 24);

		fOptionByte = buf[eOfsOptionByte];
		fPartitions = buf[eOfsPartitions];
		fBootNo = buf[eOfsBootNo];
		fPtableVersion = buf[eOfsPtableVersion];

		fImageSpaceAvailable = false;
		fImageSpaceSlots = 0;

		return ParseAdditionalPartitonSector(0, buf);
	} else {
		return false;
	}
}

bool MyidePartitionInfoLBA::ParseAdditionalPartitonSector(unsigned int secno, const unsigned char* buf)
{
	int ofs, count, part;
	if (secno > fPtableSectors) {
		return false;
	}
	if (secno == 0) {
		ofs = ePartitionOffset;
		count = 15;
		part = 0;
	} else {
		ofs = 0;
		count = 16;
		part = 15 + (secno - 1) * 16;
	}
	for (int i=0; i<count; i++) {
		int p = part + i;
		int idx = i * ePartitionBlockSize + ofs;

		unsigned char flags = buf[idx + eOfsPartitionFlags];

		fPartitionInfo[p].IsActive = flags & ePartFlagActive;
		fPartitionInfo[p].IsWritable = flags & ePartFlagWritable;

		fPartitionInfo[p].DriveNumber = buf[idx + eOfsPartitionDrive];

		fPartitionInfo[p].StartingLBA =
			 buf[idx + eOfsPartitionStart]
			 + (buf[idx + eOfsPartitionStart + 1] << 8)
			 + (buf[idx + eOfsPartitionStart + 2] << 16)
			 + (buf[idx + eOfsPartitionStart + 3] << 24);

		fPartitionInfo[p].PhysicalSectors =
			 buf[idx + eOfsPartitionSize]
			 + (buf[idx + eOfsPartitionSize + 1] << 8)
			 + (buf[idx + eOfsPartitionSize + 2] << 16)
			 + (buf[idx + eOfsPartitionSize + 3] << 24);
		fPartitionInfo[p].MaxSectors = fPartitionInfo[p].PhysicalSectors;

		unsigned char ptype = buf[idx + eOfsPartitionType];

		fPartitionInfo[p].IsEmulatedQD = false;

		switch (ptype) {
		case ePartTypeSD:
			fPartitionInfo[p].Density = eSD;
			break;
		case ePartTypeDD:
			fPartitionInfo[p].Density = eDD;
			break;
		case ePartTypeQD:
			fPartitionInfo[p].Density = eQD;
			break;
		case ePartTypeQD_emulated:
			fPartitionInfo[p].Density = eQD;
			fPartitionInfo[p].MaxSectors = fPartitionInfo[p].PhysicalSectors / 2;
			fPartitionInfo[p].IsEmulatedQD = true;
			break;
		default:
			fPartitionInfo[p].Density = eUnknownDensity;
		}
	}

	return true;
}


bool MyidePartitionInfoLBA::CreatePartitionInfo(unsigned char* buf) const
{
	memset(buf, 0, 16);

	// general stuff
	buf[eOfsTotalLBASize] = fTotalLBASectors & 0xff;
	buf[eOfsTotalLBASize + 1] = (fTotalLBASectors >> 8) & 0xff;
	buf[eOfsTotalLBASize + 2] = (fTotalLBASectors >> 16) & 0xff;
	buf[eOfsTotalLBASize + 3] = (fTotalLBASectors >> 24) & 0xff;

	buf[eOffsetDriveNumbersPartitionVersion] = eLBAPartitionTableMagic_0;
	buf[eOffsetDriveNumbersPartitionVersion+1] = eLBAPartitionTableMagic_1;
	buf[eOffsetDriveNumbersPartitionVersion+2] = eLBAPartitionTableMagic_2;
	buf[eOfsPtableVersion] = fPtableVersion;

	buf[eOfsPartitions] = fPartitions;
	buf[eOfsBootNo] = fBootNo;

	buf[eOfsOptionByte] = fOptionByte;

	buf[eOfsPtableSectors] = fPtableSectors;

	return CreateAdditionalPartitonSector(0, buf);
}

bool MyidePartitionInfoLBA::CreateAdditionalPartitonSector(unsigned int secno, unsigned char* buf) const
{
	int ofs, count, part;
	if (secno > fPtableSectors) {
		return false;
	}
	if (secno == 0) {
		ofs = ePartitionOffset;
		count = 15;
		part = 0;
		memset(buf+16, 0, 240);
	} else {
		ofs = 0;
		count = 16;
		part = 15 + (secno - 1) * 16;
		memset(buf, 0, 256);
	}
	for (int i=0; i<count; i++) {
		int p = part + i;
		int idx = i * ePartitionBlockSize + ofs;

		unsigned char flags, ptype;

		flags = 0;
		if (fPartitionInfo[p].IsActive) {
			flags |= ePartFlagActive;
		}
		if (fPartitionInfo[p].IsWritable) {
			flags |= ePartFlagWritable;
		}
		buf[idx + eOfsPartitionFlags] = flags;

/*
		if (fImageSpaceAvailable && i == eImageSpaceIdx) {
			ptype = ePartTypeImageSpace;
		} else {
*/
			switch (fPartitionInfo[p].Density) {
			case eSD:
				ptype = ePartTypeSD;
				break;
			case eDD:
				ptype = ePartTypeDD;
				break;
			case eQD:
				if (fPartitionInfo[p].IsEmulatedQD) {
					ptype = ePartTypeQD_emulated;
				} else {
					ptype = ePartTypeQD;
				}
				break;
			default:
				ptype = ePartTypeNone;
				break;
			}
/*
		}
*/
		buf[idx + eOfsPartitionStart] = fPartitionInfo[p].StartingLBA & 0xff;
		buf[idx + eOfsPartitionStart + 1] = (fPartitionInfo[p].StartingLBA >> 8) & 0xff;
		buf[idx + eOfsPartitionStart + 2] = (fPartitionInfo[p].StartingLBA >> 16) & 0xff;
		buf[idx + eOfsPartitionStart + 3] = (fPartitionInfo[p].StartingLBA >> 24) & 0xff;

		buf[idx + eOfsPartitionSize] = fPartitionInfo[p].PhysicalSectors & 0xff;
		buf[idx + eOfsPartitionSize + 1] = (fPartitionInfo[p].PhysicalSectors >> 8) & 0xff;
		buf[idx + eOfsPartitionSize + 2] = (fPartitionInfo[p].PhysicalSectors >> 16) & 0xff;
		buf[idx + eOfsPartitionSize + 3] = (fPartitionInfo[p].PhysicalSectors >> 24) & 0xff;

		buf[idx + eOfsPartitionType] = ptype;
		buf[idx + eOfsPartitionDrive] = fPartitionInfo[p].DriveNumber;
	}
	return true;
}

bool MyidePartitionInfoLBA::IsLbaPartitionTable() const
{
	return true;
}

unsigned int MyidePartitionInfoLBA::GetNumberOfPartitions() const
{
	return fMaxPartitions;
}

bool MyidePartitionInfoLBA::IsValidPartitionNumber(unsigned int partno) const
{
	return (partno >= 1) && (partno <= fMaxPartitions);
}

void MyidePartitionInfoLBA::PrintPartitionTable() const
{
	printf("\nMyIDE SDX LBA partition table V%d.%d:\n", fPtableVersion >> 4, fPtableVersion & 0xf);
        printf("total LBA sectors: %d\n", fTotalLBASectors);
/*
        printf("Image space: ");
        if (ImageSpaceEnabled()) {
                printf("start: %9d slots: %d\n",
			fPartitionInfo[eImageSpaceIdx].StartingLBA,
                        GetNumberOfImageSlots());
        } else {
                printf("disabled\n");
        }
*/
	char bin[9];
	MiscUtils::byte_to_bin(fOptionByte, bin);
        printf("partitions: %d max: %d boot: %d\n", fPartitions, fMaxPartitions, fBootNo);
        printf("option byte: %s\n", bin);

        printf("part flags drv       start dns sectors\n");
	char drno[5];
        for (unsigned int i=0;i<fMaxPartitions;i++) {
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <=15) {
			snprintf(drno, 5, "D%d:", fPartitionInfo[i].DriveNumber);
		} else {
			strcpy(drno, "----");
		}
                printf("%3d  %c %c   %-4s  %9d  %s   %5d\n", i+1,
                        fPartitionInfo[i].IsActive ? 'A' : ' ',
                        fPartitionInfo[i].IsWritable ? 'W' : ' ',
                        drno,
			fPartitionInfo[i].StartingLBA,
                        MyidePartitionInfo::DensityAsString(fPartitionInfo[i].Density, fPartitionInfo[i].IsEmulatedQD),
			fPartitionInfo[i].MaxSectors);
        }

}

bool MyidePartitionInfoLBA::IsActivePartition(unsigned int partno) const
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	return fPartitionInfo[partno-1].IsActive;
}

unsigned int MyidePartitionInfoLBA::GetPartitionSize(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return false;
	}
	if (fPartitionInfo[num-1].MaxSectors > 0) {
		return fPartitionInfo[num-1].MaxSectors - 1;
	}
	return 0;
}

EDensity MyidePartitionInfoLBA::GetPartitionDensity(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return eUnknownDensity;
	}
	return fPartitionInfo[num-1].Density;
}

bool MyidePartitionInfoLBA::IsEmulatedQDPartition(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return false;
	}
	return fPartitionInfo[num-1].IsEmulatedQD;
}

bool MyidePartitionInfoLBA::IsValidPartitionSector(unsigned int partno, unsigned int secno) const
{
	if (!IsValidPartitionNumber(partno)) {
		Assert(false);
		return false;
	}
	return (secno >= 1) && (secno < fPartitionInfo[partno-1].MaxSectors);
}

unsigned int MyidePartitionInfoLBA::GetPartitionSectorNumber(unsigned int partno, unsigned int sector) const
{
	if (fPartitionInfo[partno-1].IsEmulatedQD) {
		return fPartitionInfo[partno-1].StartingLBA + sector*2;
	} else {
		return fPartitionInfo[partno-1].StartingLBA + sector;
	}
}

// image stuff
bool MyidePartitionInfoLBA::ImageSpaceEnabled() const
{
	return fImageSpaceAvailable && 
		fPartitionInfo[eImageSpaceIdx].IsActive &&
		fPartitionInfo[eImageSpaceIdx].StartingLBA > 0 &&
		fImageSpaceSlots > 0;
}

unsigned int MyidePartitionInfoLBA::GetNumberOfImageSlots() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	return fImageSpaceSlots;
}

unsigned int MyidePartitionInfoLBA::GetSectorsPerImageSlot() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	// note: first sector is reserved for image info!
	return eSectorsPerImageSlot - 1;
}

bool MyidePartitionInfoLBA::IsValidImageSector(unsigned int secno) const
{
	// note: first sector is reserved for image info!
	return (secno >= 1) && (secno < eSectorsPerImageSlot);
}

unsigned int MyidePartitionInfoLBA::GetImageSectorNumber(unsigned int imageno, unsigned int sector) const
{
	return fPartitionInfo[eImageSpaceIdx].StartingLBA + (imageno - 1) * eSectorsPerImageSlot + sector;
}

unsigned int MyidePartitionInfoLBA::GetImageInfoSectorNumber(unsigned int imageno) const
{
	// first sector of image slot
	return fPartitionInfo[eImageSpaceIdx].StartingLBA + (imageno - 1) * eSectorsPerImageSlot;
}


// other stuff

bool MyidePartitionInfoLBA::SetPartition(unsigned int partition, bool active, bool writable, unsigned int driveno, unsigned int start, unsigned int sectors, EPartitionDensity density)
{
	if (!IsValidPartitionNumber(partition)) {
		return false;
	}
	fPartitionInfo[partition-1].IsActive = active;
	fPartitionInfo[partition-1].IsWritable = writable;
	fPartitionInfo[partition-1].DriveNumber = driveno;
	fPartitionInfo[partition-1].StartingLBA = start;

	fPartitionInfo[partition-1].PhysicalSectors = sectors;
	fPartitionInfo[partition-1].MaxSectors = sectors;

	fPartitionInfo[partition-1].IsEmulatedQD = false;
	switch (density) {
	case ePartTypeSD:
		fPartitionInfo[partition-1].Density = eSD;
		break;
	case ePartTypeDD:
		fPartitionInfo[partition-1].Density = eDD;
		break;
	case ePartTypeQD:
		fPartitionInfo[partition-1].Density = eQD;
		break;
	case ePartTypeQD_emulated:
		fPartitionInfo[partition-1].Density = eQD;
		fPartitionInfo[partition-1].IsEmulatedQD = true;
		fPartitionInfo[partition-1].MaxSectors = fPartitionInfo[partition-1].PhysicalSectors / 2;
		break;
	default:
		fPartitionInfo[partition-1].Density = eUnknownDensity;
		break;
	}

	return true;
}

bool MyidePartitionInfoLBA::SetImageSpace(bool /*enabled*/, bool /*active*/,
	unsigned int /*start*/, unsigned int /*size*/)
{
	return false;
/*
	fImageSpaceAvailable = enabled;
	if (enabled) {
		fMaxPartitions = eMaxPartitions - 1;
		fPartitionInfo[eImageSpaceIdx].IsActive = active;
		fPartitionInfo[eImageSpaceIdx].IsWritable = true;
		fPartitionInfo[eImageSpaceIdx].StartingLBA = start;
		fPartitionInfo[eImageSpaceIdx].MaxSectors = size;
		fImageSpaceSlots = size / eSectorsPerImageSlot;
	} else {
		fMaxPartitions = eMaxPartitions;
	}

	return true;
*/
}

void MyidePartitionInfoLBA::SetOptionByte(unsigned char byte)
{
	fOptionByte = byte;
}

void MyidePartitionInfoLBA::SetDiskSize(unsigned int sectors)
{
	fTotalLBASectors = sectors;
}

void MyidePartitionInfoLBA::SetPartitionsNo(unsigned char partitions)
{
	fPartitions = partitions;
}

void MyidePartitionInfoLBA::SetBootNo(unsigned char bootno)
{
	fBootNo = bootno;
}


