#ifndef MYIDEPARTITIONINFOLBA_H
#define MYIDEPARTITIONINFOLBA_H

/*
   MyidePartitionInfoLBA - infos for (new) LBA partition table

   Copyright (C) 2008-2011 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "MyidePartitionInfo.h"

class MyidePartitionInfoLBA : public MyidePartitionInfo {
public:
	MyidePartitionInfoLBA();
	virtual ~MyidePartitionInfoLBA();

	virtual bool ParsePartitionInfo(const unsigned char* buf);
	virtual bool CreatePartitionInfo(unsigned char* buf) const;

	bool ParseAdditionalPartitonSector(unsigned int secno, const unsigned char* buf);
	bool CreateAdditionalPartitonSector(unsigned int secno, unsigned char* buf) const;

	virtual bool IsLbaPartitionTable() const;

	virtual void PrintPartitionTable() const;

	// partition functions	
	virtual unsigned int GetNumberOfPartitions() const;
	virtual bool IsValidPartitionNumber(unsigned int partno) const;

	virtual bool IsActivePartition(unsigned int num) const;
	virtual unsigned int GetPartitionSize(unsigned int num) const;
	virtual EDensity GetPartitionDensity(unsigned int num) const;
	virtual bool IsEmulatedQDPartition(unsigned int num) const;

	virtual bool IsValidPartitionSector(unsigned int partno, unsigned int secno) const;

	virtual unsigned int GetPartitionSectorNumber(unsigned int partno, unsigned int sector) const;

	// image functions	
	virtual bool ImageSpaceEnabled() const;
	virtual unsigned int GetNumberOfImageSlots() const;
	virtual unsigned int GetSectorsPerImageSlot() const;

	virtual bool IsValidImageSector(unsigned int secno) const;

	// absolute
	virtual unsigned int GetImageSectorNumber(unsigned int imageno, unsigned int sector) const;
	virtual unsigned int GetImageInfoSectorNumber(unsigned int imageno) const;

	// setup partition table functions
	bool SetPartitionTableSectors(unsigned int sectors);
	inline unsigned int GetPartitionTableSectors() const;

	// partition types
	enum EPartitionDensity {
		ePartTypeNone		= 0,
		ePartTypeSD		= 0x01,		// 128 bytes/sector
		ePartTypeDD		= 0x02,		// 256 bytes/sector
		ePartTypeQD		= 0x03,		// 512 bytes/sector
		ePartTypeQD_emulated	= 0x83,		// 512 bytes/sector on 16-bit media (2 physical sectors per logical sector)
		ePartTypeImageSpace	= 0x80
	};

	bool SetPartition(unsigned int partition, bool active, bool writable, unsigned int driveno, unsigned int start, unsigned int sectors, EPartitionDensity density);
	bool SetImageSpace(bool enabled, bool active, unsigned int start, unsigned int size);
	void SetOptionByte(unsigned char byte);
	void SetDiskSize(unsigned int sectors);
	void SetPartitionsNo(unsigned char partitions);
	void SetBootNo(unsigned char bootno);

	void SetPartitionTableVersion(unsigned char ver);
	unsigned char GetPartitionTableVersion() const;

private:
	enum { eMaxPartitions = 15 };

	enum { eSectorsPerImageSlot = 1041 };

	// offsets in partition table
	enum {
		ePartitionOffset	= 0x10,		// start of partition entry #1
		eImageSpaceOffset	= 0xf0		// partition entry #15
	};

	// partition table entry #15 is either a regular partition or image space info
	enum {
		eImageSpaceIdx		= 14
	};

	// old offsets within global header block
	enum {
		eOfsTotalLBASize_old	= 0,		// 4 bytes
		eOfsLBAIdentification_old	= 4,		// must be $10
		eOfsPartitions_old	= 5,		// number of (active) partitions
		eOfsBootNo_old		= 6,		// boot partition
		eOfsOptionByte_old	= 7,		// 
		eOfsPtableSectors_old	= 8		// number of additional partition table sectors (0-15)
	};

	// new offsets within global header block
	enum {
		eOfsTotalLBASize	= 0,		// 4 bytes
		eOfsLBAIdentification	= 4,		// must be $cc, $c2, $c1
		eOfsPtableVersion	= 7,		// version number
		eOfsPtableSectors	= 8,		// number of additional partition table sectors (0-15)
		eOfsPartitions		= 9,		// number of (active) partitions
		eOfsBootNo		= 10,		// boot partition
		eOfsOptionByte		= 11		// currently unused
	};

	// maximum number of additional partition table sectors
	enum { eMaxPtableSectors = 15 };

	// partition block size
	enum {
		ePartitionBlockSize	= 0x10
	};
	// offsets within partition information block
	enum {
		eOfsPartitionStart	= 0,		// 4 bytes
		eOfsPartitionSize	= 4,		// 4 bytes
		eOfsPartitionType	= 8,		// see below
		eOfsPartitionFlags	= 9,		// see below
		eOfsPartitionDrive	= 10		// drive number (1..15)
	};

	// partition flags
	enum {
		ePartFlagActive		= 0x01,
		ePartFlagWritable	= 0x02
	};
	
        // physical drive settings
	typedef struct {
		unsigned int StartingLBA;
		unsigned int MaxSectors;
		unsigned int PhysicalSectors;
		EDensity Density;
		unsigned int DriveNumber;
		bool IsActive;
		bool IsWritable;
		bool IsEmulatedQD;
	} LBAPartition;
        LBAPartition *fPartitionInfo;

	unsigned int fTotalLBASectors;

	bool fImageSpaceAvailable;
	unsigned int fImageSpaceSlots;

	unsigned int fMaxPartitions;
	unsigned int fPartitions;
	unsigned int fBootNo;
	unsigned int fPtableSectors;
	unsigned char fPtableVersion;
};

inline unsigned int MyidePartitionInfoLBA::GetPartitionTableSectors() const
{
	return fPtableSectors;
}

inline void MyidePartitionInfoLBA::SetPartitionTableVersion(unsigned char ver)
{
	fPtableVersion = ver;
}

inline unsigned char MyidePartitionInfoLBA::GetPartitionTableVersion() const
{
	return fPtableVersion;
}


#endif
